const {truncate} = require('lodash');
const striptags = require('striptags');
const {MetricsFetcher, METRICS} = require('./ga-fetcher');

// recursively get text from all nested children to form a single string of text
// replace all whitespace of each child.value with a single space
// want to join inner elements with '' and outer elements with a space
const getChildrenText = ({children}) =>
  Array.isArray(children)
    ? children
        .map(child =>
          ['text', 'inlineCode'].includes(child.type)
            ? child.value
            : getChildrenText(child)
        )
        .join('')
    : '';

// make an exception for <Void> in some android docs
const ALLOWED_TAGS = ['void'];

const reduceMdxHeadings = (acc, {items = [], url, title}) => {
  const stripped = striptags(title, ALLOWED_TAGS);
  return items.reduce(reduceMdxHeadings, {
    ...acc,
    [stripped]: Object.prototype.hasOwnProperty.call(acc, stripped)
      ? [...acc[stripped], url]
      : [url]
  });
};

const getMdxHeading = (child, headings) => {
  if (child.type === 'heading') {
    // get text children from heading
    const title = getChildrenText(child);

    // want to strip any tags that might be in title to match values in headings
    // ex. client docs have a heading `MutationTuple<TData, TVariables> result tuple`
    // where the title includes the tags, but the corresponding value in headings is `MutationTuple result tuple`
    const strippedTitle = striptags(title, ALLOWED_TAGS);
    // splice the first result for that title in the headings object
    // this lets us account for multiple headings with the same title
    // but different URLs in the same page
    const [hash] = headings[strippedTitle].splice(0, 1);

    return {
      title,
      hash,
      depth: child.depth
    };
  }
};

// takes an array of top-level elements as children and chunks them up into
// records for each section beginning and ending on a heading
const createRecords = ({
  children,
  getHeading = getMdxHeading,
  getText = getChildrenText,
  tableOfContents,
  id,
  url,
  otherProperties
}) => {
  // create a mapping of heading title -> url for MDX pages
  const mdxHeadings =
    tableOfContents && // tableOfContents may not exist (WP)
    tableOfContents.items && // or may not have an items property (MD)
    tableOfContents.items.reduce(reduceMdxHeadings, {});

  const sections = children
    .reduce(
      (acc, child) => {
        // these will return a heading object if the child is a heading
        const heading = getHeading(child, mdxHeadings);

        if (heading) {
          // determine the heading's ancestors by looping through existing
          // sections and comparing their depth with the current heading's
          let ancestors = [];
          for (const section of acc) {
            if (section.depth < heading.depth) {
              ancestors = [
                {
                  title: section.title,
                  url: url + section.hash
                },
                ...section.ancestors
              ];
              break;
            }
          }

          return [
            {
              ...heading,
              children: [],
              ancestors
            },
            ...acc
          ];
        }

        // if the child is not a heading, add it to the section as a child
        acc[0].children.push(child);
        return acc;
      },
      [{children: []}]
    )
    .reverse();

  return sections
    .map((section, index) => {
      const {title: sectionTitle, hash, children, ancestors = []} = section;

      const text = children
        .map(getText)
        .join(' ') // separate 'nodes' with a space
        .replace(/\s+/g, ' ') // remove any duplicate spaces
        .trim();

      return {
        objectID: `${id}_${index}`,
        index,
        sectionTitle,
        text,
        excerpt: truncate(text, {length: 100, separator: ' '}),
        ancestors,
        url: hash ? url + hash : url,
        ...otherProperties
      };
    })
    .filter(section => section.text); // we don't add records that have no text
};

module.exports = {
  algoliaSettings: require('./settings'),
  MetricsFetcher,
  METRICS,
  getMdxHeading,
  createRecords,
  getChildrenText
};
